@@ -18,6 +18,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <fcntl.h>
 
 #include <openssl/bio.h>
 #include <openssl/err.h>
@@ -32,6 +33,10 @@
 #include "https.h"
 #include "match.h"
 
+#ifdef HAVE_X509_TEA_SET_STATE
+extern void X509_TEA_set_state(int change);
+#endif
+
 struct https_ctx {
         SSL_CTX              *ssl_ctx;
         char	             *ikey;
@@ -104,7 +109,6 @@ _SSL_check_server_cert(SSL *ssl, const char *hostname)
         ASN1_STRING *tmp;
         int i, n, match = -1;
         const char *p;
-        
         if (SSL_get_verify_mode(ssl) == SSL_VERIFY_NONE ||
             (cert = SSL_get_peer_certificate(ssl)) == NULL) {
                 return (1);
@@ -146,10 +150,41 @@ _SSL_check_server_cert(SSL *ssl, const char *hostname)
         return (match > 0);
 }
 
-// Return -1 on hard error (abort), 0 on timeout, >= 1 on successful wakeup
+/* Wait msecs milliseconds for the fd to become writable.  Return
+ * -1 on error, 0 on timeout, and >0 if the fd is writable.
+ */
+static int
+_fd_wait(int fd, int msecs)
+{
+    struct pollfd pfd;
+    int result;
+
+    pfd.fd = fd;
+    pfd.events = POLLOUT | POLLWRBAND;
+    pfd.revents = 0;
+
+    if (msecs < 0) {
+        msecs = -1;
+    }
+
+    do {
+        result = poll(&pfd, 1, msecs);
+    } while (result == -1 && errno == EINTR);
+
+    if (result <= 0) {
+        return result;
+    }
+    if (pfd.revents & POLLERR) {
+        return -1;
+    }
+    return (pfd.revents & pfd.events ? 1 : -1);
+}
+
+/* Return -1 on hard error (abort), 0 on timeout, >= 1 on successful wakeup */
 static int
 _BIO_wait(BIO *cbio, int msecs)
 {
+        int result;
         if (!BIO_should_retry(cbio)) {
                 return (-1);
         }
@@ -174,14 +209,19 @@ _BIO_wait(BIO *cbio, int msecs)
                accept any negative value. */
             msecs = -1;
         }
-        int result = poll(&pfd, 1, msecs);
+        do {
+            result = poll(&pfd, 1, msecs);
+        } while (result == -1 && errno == EINTR);
 
-        // Timeout or poll internal error
+        /* Timeout or poll internal error */
         if (result <= 0) {
                 return (result);
         }
+        if (pfd.revents & POLLERR) {
+            return -1;
+        }
 
-        // Return 1 if the event was not an error
+        /* Return 1 if the event was not an error */
         return (pfd.revents & pfd.events ? 1 : -1);
 }
 
@@ -195,6 +235,117 @@ _BIO_new_base64(void)
         return (b64);
 }
 
+/*
+ * Establishes the connection to the Duo server.  On successful return,
+ * req->cbio is connected and ready to use.
+ * Return HTTPS_OK on success, error code on failure.
+ */
+static HTTPScode
+_establish_connection(struct https_request * const req,
+        const char * const api_host,
+        const char * const api_port)
+{
+#ifndef HAVE_GETADDRINFO
+    /* Systems that don't have getaddrinfo can use the BIO
+       wrappers, but only get IPv4 support. */
+    int n;
+
+    if ((req->cbio = BIO_new(BIO_s_connect())) == NULL) {
+        ctx->errstr = _SSL_strerror();
+        return HTTPS_ERR_LIB;
+    }
+    BIO_set_conn_hostname(req->cbio, api_host);
+    BIO_set_conn_port(req->cbio, api_port);
+    BIO_set_nbio(req->cbio, 1);
+
+    while (BIO_do_connect(req->cbio) <= 0) {
+        if ((n = _BIO_wait(req->cbio, 10000)) != 1) {
+            ctx->errstr = n ? _SSL_strerror() :
+                "Connection timed out";
+            return (n ? HTTPS_ERR_SYSTEM : HTTPS_ERR_SERVER);
+        }
+    }
+
+    return HTTPS_OK;
+
+#else /* HAVE_GETADDRINFO */
+
+    /* IPv6 Support
+     * BIO wrapped io does not support IPv6 addressing.  To work around,
+     * resolve the address and connect the socket manually.  Then pass
+     * the connected socket to the BIO wrapper with BIO_new_socket.
+     */
+    int connected_socket = -1;
+    int socket_error = 0;
+    /* Address Lookup */
+    struct addrinfo *res = NULL;
+    struct addrinfo *cur_res = NULL;
+    struct addrinfo hints;
+    int error;
+
+    memset(&hints, 0, sizeof(hints));
+    hints.ai_family = PF_UNSPEC;
+    hints.ai_socktype = SOCK_STREAM;
+
+    error = getaddrinfo(api_host,
+            api_port,
+            &hints,
+            &res);
+    if (error) {
+        ctx->errstr = gai_strerror(error);
+        return HTTPS_ERR_SYSTEM;
+    }
+
+    /* Connect */
+    for (cur_res = res; cur_res; cur_res = cur_res->ai_next) {
+        int sock_flags;
+
+        connected_socket = socket(cur_res->ai_family,
+                cur_res->ai_socktype,
+                cur_res->ai_protocol);
+        if (connected_socket == -1) {
+            continue;
+        }
+        sock_flags = fcntl(connected_socket, F_GETFL, 0);
+        fcntl(connected_socket, F_SETFL, sock_flags|O_NONBLOCK);
+
+        if (connect(connected_socket, cur_res->ai_addr, cur_res->ai_addrlen) != 0 &&
+                errno != EINPROGRESS) {
+            close(connected_socket);
+            connected_socket = -1;
+            continue;
+        }
+        socket_error = _fd_wait(connected_socket, 10000);
+        if (socket_error != 1) {
+            close(connected_socket);
+            connected_socket = -1;
+            continue;
+        }
+        /* Connected! */
+        break;
+    }
+    cur_res = NULL;
+    freeaddrinfo(res);
+    res = NULL;
+
+    if (connected_socket == -1) {
+        ctx->errstr = "Failed to connect";
+        return socket_error ? HTTPS_ERR_SYSTEM : HTTPS_ERR_SERVER;
+    }
+
+    if ((req->cbio = BIO_new_socket(connected_socket, BIO_CLOSE)) == NULL) {
+        ctx->errstr = _SSL_strerror();
+        return (HTTPS_ERR_LIB);
+    }
+    BIO_set_conn_hostname(req->cbio, api_host);
+    BIO_set_conn_port(req->cbio, api_port);
+    BIO_set_nbio(req->cbio, 1);
+
+    return HTTPS_OK;
+
+#endif /* HAVE_GETADDRINFO */
+}
+
 HTTPScode
 https_init(const char *ikey, const char *skey,
     const char *useragent, const char *cafile)
@@ -203,7 +354,7 @@ https_init(const char *ikey, const char *skey,
         X509 *cert;
         BIO *bio;
         char *p;
-        
+
         if ((ctx = calloc(1, sizeof(*ctx))) == NULL ||
             (ctx->ikey = strdup(ikey)) == NULL ||
             (ctx->skey = strdup(skey)) == NULL ||
@@ -212,6 +363,13 @@ https_init(const char *ikey, const char *skey,
                 return (HTTPS_ERR_SYSTEM);
         }
         /* Initialize SSL context */
+#ifdef HAVE_X509_TEA_SET_STATE
+        /* If applicable, disable use of Apple's Trust Evaluation Agent for certificate
+         * validation, to enforce proper CA pinning:
+         * http://www.opensource.apple.com/source/OpenSSL098/OpenSSL098-35.1/src/crypto/x509/x509_vfy_apple.h
+         */
+        X509_TEA_set_state(0);
+#endif
         SSL_library_init();
         SSL_load_error_strings();
         OpenSSL_add_all_algorithms();
@@ -297,7 +455,10 @@ https_open(struct https_request **reqp, const char *host)
         BIO *b64, *sbio;
         char *p;
         int n;
+        int connection_error = 0;
 
+        const char *api_host;
+        const char *api_port;
         /* Set up our handle */
         n = 1;
         if ((req = calloc(1, sizeof(*req))) == NULL ||
@@ -313,8 +474,7 @@ https_open(struct https_request **reqp, const char *host)
         } else {
                 req->port = "443";
         }
-        if ((req->cbio = BIO_new(BIO_s_connect())) == NULL ||
-            (req->body = BIO_new(BIO_s_mem())) == NULL) {
+        if ((req->body = BIO_new(BIO_s_mem())) == NULL) {
                 ctx->errstr = _SSL_strerror();
                 https_close(&req);
                 return (HTTPS_ERR_LIB);
@@ -324,22 +484,19 @@ https_open(struct https_request **reqp, const char *host)
 
         /* Connect to server */
         if (ctx->proxy) {
-                BIO_set_conn_hostname(req->cbio, ctx->proxy);
-                BIO_set_conn_port(req->cbio, ctx->proxy_port);
+            api_host = ctx->proxy;
+            api_port = ctx->proxy_port;
         } else {
-                BIO_set_conn_hostname(req->cbio, req->host);
-                BIO_set_conn_port(req->cbio, req->port);
+            api_host = req->host;
+            api_port = req->port;
         }
-        BIO_set_nbio(req->cbio, 1);
-        
-        while (BIO_do_connect(req->cbio) <= 0) {
-                if ((n = _BIO_wait(req->cbio, 10000)) != 1) {
-                        ctx->errstr = n ? _SSL_strerror() :
-                            "Connection timed out";
-                        https_close(&req);
-                        return (n ? HTTPS_ERR_SYSTEM : HTTPS_ERR_SERVER);
-                }
+
+        connection_error = _establish_connection(req, api_host, api_port);
+        if (connection_error != HTTPS_OK) {
+            https_close(&req);
+            return connection_error;
         }
+
         /* Tunnel through proxy, if specified */
         if (ctx->proxy != NULL) {
                 BIO_printf(req->cbio,
@@ -364,7 +521,15 @@ https_open(struct https_request **reqp, const char *host)
                 
                 while ((n = BIO_read(req->cbio, ctx->parse_buf,
                             sizeof(ctx->parse_buf))) <= 0) {
-                        _BIO_wait(req->cbio, 5000);
+                    if ((n = _BIO_wait(req->cbio, 5000)) != 1) {
+                        if (n == 0) {
+                            ctx->errstr = "Proxy connection timed out";
+                        } else {
+                            ctx->errstr = "Proxy connection error";
+                        }
+                        https_close(&req);
+                        return HTTPS_ERR_SERVER;
+                    }
                 }
                 /* Tolerate HTTP proxies that respond with an
                    incorrect HTTP version number */
@@ -384,9 +549,19 @@ https_open(struct https_request **reqp, const char *host)
                 https_close(&req);
                 return (HTTPS_ERR_LIB);
         }
+
         req->cbio = BIO_push(sbio, req->cbio);
         BIO_get_ssl(req->cbio, &req->ssl);
-        
+
+#ifdef SSL_CTRL_SET_TLSEXT_HOSTNAME
+        /* Enable SNI support */
+        if ((n = SSL_set_tlsext_host_name(req->ssl, api_host)) != 1) {
+            ctx->errstr = "Setting SNI failed";
+            https_close(&req);
+            return (HTTPS_ERR_LIB);
+        }
+#endif
+
         while (BIO_do_handshake(req->cbio) <= 0) {
                 if ((n = _BIO_wait(req->cbio, 5000)) != 1) {
                         ctx->errstr = n ? _SSL_strerror() :
@@ -517,7 +692,8 @@ https_send(struct https_request *req, const char *method, const char *uri,
 }
 
 HTTPScode
-https_recv(struct https_request *req, int *code, const char **body, int *len)
+https_recv(struct https_request *req, int *code, const char **body, int *len,
+        int msecs)
 {
         int n, err;
         
@@ -529,7 +705,7 @@ https_recv(struct https_request *req, int *code, const char **body, int *len)
         while (!req->done) {
                 while ((n = BIO_read(req->cbio, ctx->parse_buf,
                             sizeof(ctx->parse_buf))) <= 0) {
-                        if ((n = _BIO_wait(req->cbio, -1)) != 1) {
+                        if ((n = _BIO_wait(req->cbio, msecs)) != 1) {
                                 ctx->errstr = n ? _SSL_strerror() :
                                     "Connection closed";
                                 return (HTTPS_ERR_SERVER);
