@@ -27,6 +27,7 @@
 
 #include "util.h"
 #include "duo.h"
+#include "shell.h"
 
 #ifndef DUO_PRIVSEP_USER
 #define DUO_PRIVSEP_USER    "duo"
@@ -109,6 +110,19 @@ _print_motd()
     return (0);
 }
 
+static void
+restore_http_proxy(const char* http_proxy)
+{
+    if (http_proxy != NULL)
+    {
+        setenv("http_proxy", http_proxy, 1);
+    }
+    else
+    {
+        unsetenv("http_proxy");
+    }
+}
+
 static int
 do_auth(struct login_ctx *ctx, const char *cmd)
 {
@@ -118,7 +132,7 @@ do_auth(struct login_ctx *ctx, const char *cmd)
     duo_t *duo;
     duo_code_t code;
     const char *config, *p, *duouser;
-    const char *ip, *host = NULL;
+    const char *ip, *host, *orig_http_proxy = NULL;
     char buf[64];
     int i, flags, ret, prompts, matched;
     int headless = 0;
@@ -178,11 +192,14 @@ do_auth(struct login_ctx *ctx, const char *cmd)
             ip = strtok(buf, " ");
             host = ip;
         } else {
-            ip = (cfg.local_ip_fallback ? duo_local_ip() : NULL);
+            if (cfg.local_ip_fallback) {
+                host = duo_local_ip();
+            }
         }
     }
 
     /* Honor configured http_proxy */
+    orig_http_proxy = getenv("http_proxy");
     if (cfg.http_proxy != NULL) {
         setenv("http_proxy", cfg.http_proxy, 1);
     }
@@ -190,9 +207,11 @@ do_auth(struct login_ctx *ctx, const char *cmd)
     /* Try Duo auth. */
     if ((duo = duo_open(cfg.apihost, cfg.ikey, cfg.skey,
                     "login_duo/" PACKAGE_VERSION,
-                    cfg.noverify ? "" : cfg.cafile)) == NULL) {
+                    cfg.noverify ? "" : cfg.cafile,
+                    cfg.https_timeout)) == NULL) {
         duo_log(LOG_ERR, "Couldn't open Duo API handle",
             pw->pw_name, host, NULL);
+        restore_http_proxy(orig_http_proxy);
         return (EXIT_FAILURE);
     }
 
@@ -263,6 +282,7 @@ do_auth(struct login_ctx *ctx, const char *cmd)
     }
     duo_close(duo);
 
+    restore_http_proxy(orig_http_proxy);
     return (ret);
 }
 
@@ -295,26 +315,34 @@ do_exec(struct login_ctx *ctx, const char *cmd)
         struct passwd *pw;
     const char *shell0;
     char argv0[256];
+    const char *user_shell;
     int n;
 
         if ((pw = getpwuid(ctx->uid)) == NULL)
                 die("Who are you?");
-        
-    if ((shell0 = strrchr(pw->pw_shell, '/')) != NULL) {
+
+    /* Check to see if we have a shell from getpwuid() */
+    if (NULL == pw->pw_shell) {
+      user_shell = _DEFAULT_SHELL; /* No shell so use the default. */
+    }
+    else {
+      user_shell = pw->pw_shell; /* Use the shell provided by getpwuid() */
+    }
+    if ((shell0 = strrchr(user_shell, '/')) != NULL) {
         shell0++;
     } else {
-        shell0 = pw->pw_shell;
+        shell0 = user_shell;
     }
     if (cmd != NULL) {
-        execl(pw->pw_shell, shell0, "-c", cmd, (char *)NULL);
+        execl(user_shell, shell0, "-c", cmd, (char *)NULL);
     } else {
         n = snprintf(argv0, sizeof(argv0), "-%s", shell0);
         if (n == -1 || n >= sizeof(argv0)) {
-            die("%s: Invalid argument", pw->pw_shell);
+            die("%s: Invalid argument", user_shell);
         }
-        execl(pw->pw_shell, argv0, (char *)NULL);
+        execl(user_shell, argv0, (char *)NULL);
     }
-    die("%s: %s", pw->pw_shell, strerror(errno));
+    die("%s: %s", user_shell, strerror(errno));
 }
 
 static char *
@@ -350,6 +378,7 @@ main(int argc, char *argv[])
     struct passwd *pw;
     pid_t pid;
     int c, stat;
+    pid_t wait_res;
     
     memset(ctx, 0, sizeof(ctx));
     
@@ -401,7 +430,11 @@ main(int argc, char *argv[])
                     strerror(errno));
             }
             /* Check auth child status. */
-            if (waitpid(pid, &stat, 0) != pid) {
+            while ((wait_res = waitpid(pid, &stat, 0)) == -1 &&
+                    errno == EINTR) {
+                ;
+            }
+            if (wait_res != pid) {
                 die("waitpid: %s", strerror(errno));
             }
             if (WEXITSTATUS(stat) == 0) {
