@@ -90,10 +90,10 @@ __duo_status(void *arg, const char *msg)
 static char *
 __duo_prompt(void *arg, const char *prompt, char *buf, size_t bufsz)
 {
-	char *p;
-	
+	char *p = NULL;
+
 	if (pam_prompt((pam_handle_t *)arg, PAM_PROMPT_ECHO_ON, &p,
-		"%s", prompt) != PAM_SUCCESS) {
+		"%s", prompt) != PAM_SUCCESS || p == NULL) {
 		return (NULL);
 	}
 	strlcpy(buf, p, bufsz);
@@ -101,6 +101,19 @@ __duo_prompt(void *arg, const char *prompt, char *buf, size_t bufsz)
 	return (buf);
 }
 
+static void
+restore_http_proxy(const char* http_proxy)
+{
+	if (http_proxy != NULL)
+	{
+		setenv("http_proxy", http_proxy, 1);
+	}
+	else
+	{
+		unsetenv("http_proxy");
+	}
+}
+
 PAM_EXTERN int
 pam_sm_authenticate(pam_handle_t *pamh, int pam_flags,
     int argc, const char *argv[])
@@ -110,8 +123,17 @@ pam_sm_authenticate(pam_handle_t *pamh, int pam_flags,
 	struct in_addr addr;
 	duo_t *duo;
 	duo_code_t code;
-	duopam_const char *config, *cmd, *p, *service, *user;
-	const char *ip, *host;
+
+	/*
+	 * Only variables that will be passed to a pam_* function
+	 * need to be marked as 'duopam_const char *', anything else
+	 * should be 'const char *'. This is because there are different
+	 * PAM implementations, some with the const qualifier, and some
+	 * without.
+	 */
+	duopam_const char *ip, *service, *user;
+	const char *cmd, *p, *config, *host, *orig_http_proxy;
+
 	int i, flags, pam_err, matched;
 
 	duo_config_default(&cfg);
@@ -168,7 +190,7 @@ pam_sm_authenticate(pam_handle_t *pamh, int pam_flags,
             flags |= DUO_FLAG_SYNC;
     } else if (strcmp(service, "sudo") == 0) {
             cmd = getenv("SUDO_COMMAND");
-    } else if (strcmp(service, "su") == 0) {
+    } else if (strcmp(service, "su") == 0 || strcmp(service, "su-l") == 0) {
             /* Check calling user for Duo auth, just like sudo */
             if ((pw = getpwuid(getuid())) == NULL) {
                     return (PAM_USER_UNKNOWN);
@@ -183,6 +205,15 @@ pam_sm_authenticate(pam_handle_t *pamh, int pam_flags,
         return (PAM_SUCCESS);
     }
 
+    /* Use GECOS field if called for */
+    if (cfg.send_gecos) {
+      if (strlen(pw->pw_gecos) > 0) {
+          user = pw->pw_gecos;
+      } else {
+          duo_log(LOG_WARNING, "Empty GECOS field", pw->pw_name, NULL, NULL);
+      }
+    }
+
     /* Grab the remote host */
 	ip = NULL;
 	pam_get_item(pamh, PAM_RHOST,
@@ -194,10 +225,13 @@ pam_sm_authenticate(pam_handle_t *pamh, int pam_flags,
 	}
 	if (!inet_aton(ip, &addr)) {
 		/* We have a hostname, don't try to resolve, check fallback */
-		ip = (cfg.local_ip_fallback ? duo_local_ip() : NULL);
+		if (cfg.local_ip_fallback) {
+			host = duo_local_ip();
+		}
 	}
 
 	/* Honor configured http_proxy */
+	orig_http_proxy = getenv("http_proxy");
 	if (cfg.http_proxy != NULL) {
 		setenv("http_proxy", cfg.http_proxy, 1);
 	}
@@ -205,8 +239,9 @@ pam_sm_authenticate(pam_handle_t *pamh, int pam_flags,
 	/* Try Duo auth */
 	if ((duo = duo_open(cfg.apihost, cfg.ikey, cfg.skey,
                     "pam_duo/" PACKAGE_VERSION,
-                    cfg.noverify ? "" : cfg.cafile)) == NULL) {
-		duo_log(LOG_ERR, "Couldn't open Duo API handle", user, host, NULL);
+                    cfg.noverify ? "" : cfg.cafile, cfg.https_timeout)) == NULL) {
+		duo_log(LOG_ERR, "Couldn't open Duo API handle", pw->pw_name, host, NULL);
+		restore_http_proxy(orig_http_proxy);
 		return (PAM_SERVICE_ERR);
 	}
 	duo_set_conv_funcs(duo, __duo_prompt, __duo_status, pamh);
@@ -216,13 +251,13 @@ pam_sm_authenticate(pam_handle_t *pamh, int pam_flags,
 	}
 
 	pam_err = PAM_SERVICE_ERR;
-	
+
 	for (i = 0; i < cfg.prompts; i++) {
 		code = duo_login(duo, user, host, flags,
                     cfg.pushinfo ? cmd : NULL);
 		if (code == DUO_FAIL) {
 			duo_log(LOG_WARNING, "Failed Duo login",
-			    user, host, duo_geterr(duo));
+			    pw->pw_name, host, duo_geterr(duo));
 			if ((flags & DUO_FLAG_SYNC) == 0) {
 				pam_info(pamh, "%s", "");
                         }
@@ -233,25 +268,25 @@ pam_sm_authenticate(pam_handle_t *pamh, int pam_flags,
 		if (code == DUO_OK) {
 			if ((p = duo_geterr(duo)) != NULL) {
 				duo_log(LOG_WARNING, "Skipped Duo login",
-				    user, host, p);
+				    pw->pw_name, host, p);
 			} else {
 				duo_log(LOG_INFO, "Successful Duo login",
-				    user, host, NULL);
+				    pw->pw_name, host, NULL);
 			}
 			pam_err = PAM_SUCCESS;
 		} else if (code == DUO_ABORT) {
 			duo_log(LOG_WARNING, "Aborted Duo login",
-			    user, host, duo_geterr(duo));
+			    pw->pw_name, host, duo_geterr(duo));
 			pam_err = PAM_ABORT;
 		} else if (cfg.failmode == DUO_FAIL_SAFE &&
                     (code == DUO_CONN_ERROR ||
                      code == DUO_CLIENT_ERROR || code == DUO_SERVER_ERROR)) {
 			duo_log(LOG_WARNING, "Failsafe Duo login",
-			    user, host, duo_geterr(duo));
+			    pw->pw_name, host, duo_geterr(duo));
 			pam_err = PAM_SUCCESS;
 		} else {
 			duo_log(LOG_ERR, "Error in Duo login",
-			    user, host, duo_geterr(duo));
+			    pw->pw_name, host, duo_geterr(duo));
 			pam_err = PAM_SERVICE_ERR;
 		}
 		break;
@@ -261,6 +296,7 @@ pam_sm_authenticate(pam_handle_t *pamh, int pam_flags,
 	}
 	duo_close(duo);
 	
+	restore_http_proxy(orig_http_proxy);
 	return (pam_err);
 }
 
